# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Abstract base class FederationManager."""


from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from flwr.common.typing import Federation

if TYPE_CHECKING:
    from flwr.server.superlink.linkstate.linkstate import LinkState


class FederationManager(ABC):
    """Abstract base class for FederationManager."""

    @property
    def linkstate(self) -> "LinkState":
        """Return the LinkState instance."""
        if not (ret := getattr(self, "_linkstate", None)):
            raise RuntimeError("linkstate not set. Assign to linkstate property first.")
        return ret  # type: ignore

    @linkstate.setter
    def linkstate(self, linkstate: "LinkState") -> None:
        """Set the LinkState instance."""
        self._linkstate = linkstate

    @abstractmethod
    def exists(self, federation: str) -> bool:
        """Check if a federation exists."""

    @abstractmethod
    def has_member(self, flwr_aid: str, federation: str) -> bool:
        """Check if the given account is a member of the federation."""

    @abstractmethod
    def filter_nodes(self, node_ids: set[int], federation: str) -> set[int]:
        """Given a list of node IDs, return sublist with nodes in federation."""

    @abstractmethod
    def has_node(self, node_id: int, federation: str) -> bool:
        """Given a node ID, check if it is in the federation."""

    @abstractmethod
    def get_federations(self, flwr_aid: str) -> list[str]:
        """Get federations of which the account is a member."""

    @abstractmethod
    def get_details(self, federation: str) -> Federation:
        """Get details of the federation."""



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Concrete NoOp implementation for Servicer-side account authentication and
authorization plugins."""


from collections.abc import Sequence
from pathlib import Path

from flwr.common.constant import NOOP_ACCOUNT_NAME, NOOP_FLWR_AID, AuthnType
from flwr.common.typing import (
    AccountAuthCredentials,
    AccountAuthLoginDetails,
    AccountInfo,
)

from .auth_plugin import ControlAuthnPlugin, ControlAuthzPlugin

NOOP_ACCOUNT_INFO = AccountInfo(
    flwr_aid=NOOP_FLWR_AID,
    account_name=NOOP_ACCOUNT_NAME,
)


class NoOpControlAuthnPlugin(ControlAuthnPlugin):
    """No-operation implementation of ControlAuthnPlugin."""

    def __init__(
        self,
        account_auth_config_path: Path,
        verify_tls_cert: bool,
    ):
        pass

    def get_login_details(self) -> AccountAuthLoginDetails | None:
        """Get the login details."""
        # This allows the `flwr login` command to load the NoOp plugin accordingly,
        # which then raises a LoginError when attempting to login.
        return AccountAuthLoginDetails(
            authn_type=AuthnType.NOOP,  # No operation authn type
            device_code="",
            verification_uri_complete="",
            expires_in=0,
            interval=0,
        )

    def validate_tokens_in_metadata(
        self, metadata: Sequence[tuple[str, str | bytes]]
    ) -> tuple[bool, AccountInfo | None]:
        """Return valid for no-op plugin."""
        return True, NOOP_ACCOUNT_INFO

    def get_auth_tokens(self, device_code: str) -> AccountAuthCredentials | None:
        """Get authentication tokens."""
        raise RuntimeError("NoOp plugin does not support getting auth tokens.")

    def refresh_tokens(
        self, metadata: Sequence[tuple[str, str | bytes]]
    ) -> tuple[Sequence[tuple[str, str | bytes]] | None, AccountInfo | None]:
        """Refresh authentication tokens in the provided metadata."""
        return metadata, NOOP_ACCOUNT_INFO


class NoOpControlAuthzPlugin(ControlAuthzPlugin):
    """No-operation implementation of ControlAuthzPlugin."""

    def __init__(self, account_auth_config_path: Path, verify_tls_cert: bool):
        pass

    def authorize(self, account_info: AccountInfo) -> bool:
        """Return True for no-op plugin."""
        return True



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: flwr/proto/run.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'flwr/proto/run.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from flwr.proto import node_pb2 as flwr_dot_proto_dot_node__pb2
from flwr.proto import recorddict_pb2 as flwr_dot_proto_dot_recorddict__pb2
from flwr.proto import transport_pb2 as flwr_dot_proto_dot_transport__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x66lwr/proto/run.proto\x12\nflwr.proto\x1a\x15\x66lwr/proto/node.proto\x1a\x1b\x66lwr/proto/recorddict.proto\x1a\x1a\x66lwr/proto/transport.proto\"\xf4\x02\n\x03Run\x12\x0e\n\x06run_id\x18\x01 \x01(\x04\x12\x0e\n\x06\x66\x61\x62_id\x18\x02 \x01(\t\x12\x13\n\x0b\x66\x61\x62_version\x18\x03 \x01(\t\x12<\n\x0foverride_config\x18\x04 \x03(\x0b\x32#.flwr.proto.Run.OverrideConfigEntry\x12\x10\n\x08\x66\x61\x62_hash\x18\x05 \x01(\t\x12\x12\n\npending_at\x18\x06 \x01(\t\x12\x13\n\x0bstarting_at\x18\x07 \x01(\t\x12\x12\n\nrunning_at\x18\x08 \x01(\t\x12\x13\n\x0b\x66inished_at\x18\t \x01(\t\x12%\n\x06status\x18\n \x01(\x0b\x32\x15.flwr.proto.RunStatus\x12\x10\n\x08\x66lwr_aid\x18\x0b \x01(\t\x12\x12\n\nfederation\x18\x0c \x01(\t\x1aI\n\x13OverrideConfigEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12!\n\x05value\x18\x02 \x01(\x0b\x32\x12.flwr.proto.Scalar:\x02\x38\x01\"@\n\tRunStatus\x12\x0e\n\x06status\x18\x01 \x01(\t\x12\x12\n\nsub_status\x18\x02 \x01(\t\x12\x0f\n\x07\x64\x65tails\x18\x03 \x01(\t\"?\n\rGetRunRequest\x12\x1e\n\x04node\x18\x01 \x01(\x0b\x32\x10.flwr.proto.Node\x12\x0e\n\x06run_id\x18\x02 \x01(\x04\".\n\x0eGetRunResponse\x12\x1c\n\x03run\x18\x01 \x01(\x0b\x32\x0f.flwr.proto.Run\"S\n\x16UpdateRunStatusRequest\x12\x0e\n\x06run_id\x18\x01 \x01(\x04\x12)\n\nrun_status\x18\x02 \x01(\x0b\x32\x15.flwr.proto.RunStatus\"\x19\n\x17UpdateRunStatusResponse\"F\n\x13GetRunStatusRequest\x12\x1e\n\x04node\x18\x01 \x01(\x0b\x32\x10.flwr.proto.Node\x12\x0f\n\x07run_ids\x18\x02 \x03(\x04\"\xb1\x01\n\x14GetRunStatusResponse\x12L\n\x0frun_status_dict\x18\x01 \x03(\x0b\x32\x33.flwr.proto.GetRunStatusResponse.RunStatusDictEntry\x1aK\n\x12RunStatusDictEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.flwr.proto.RunStatus:\x02\x38\x01\"-\n\x1bGetFederationOptionsRequest\x12\x0e\n\x06run_id\x18\x01 \x01(\x04\"T\n\x1cGetFederationOptionsResponse\x12\x34\n\x12\x66\x65\x64\x65ration_options\x18\x01 \x01(\x0b\x32\x18.flwr.proto.ConfigRecordb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'flwr.proto.run_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_RUN_OVERRIDECONFIGENTRY']._loaded_options = None
  _globals['_RUN_OVERRIDECONFIGENTRY']._serialized_options = b'8\001'
  _globals['_GETRUNSTATUSRESPONSE_RUNSTATUSDICTENTRY']._loaded_options = None
  _globals['_GETRUNSTATUSRESPONSE_RUNSTATUSDICTENTRY']._serialized_options = b'8\001'
  _globals['_RUN']._serialized_start=117
  _globals['_RUN']._serialized_end=489
  _globals['_RUN_OVERRIDECONFIGENTRY']._serialized_start=416
  _globals['_RUN_OVERRIDECONFIGENTRY']._serialized_end=489
  _globals['_RUNSTATUS']._serialized_start=491
  _globals['_RUNSTATUS']._serialized_end=555
  _globals['_GETRUNREQUEST']._serialized_start=557
  _globals['_GETRUNREQUEST']._serialized_end=620
  _globals['_GETRUNRESPONSE']._serialized_start=622
  _globals['_GETRUNRESPONSE']._serialized_end=668
  _globals['_UPDATERUNSTATUSREQUEST']._serialized_start=670
  _globals['_UPDATERUNSTATUSREQUEST']._serialized_end=753
  _globals['_UPDATERUNSTATUSRESPONSE']._serialized_start=755
  _globals['_UPDATERUNSTATUSRESPONSE']._serialized_end=780
  _globals['_GETRUNSTATUSREQUEST']._serialized_start=782
  _globals['_GETRUNSTATUSREQUEST']._serialized_end=852
  _globals['_GETRUNSTATUSRESPONSE']._serialized_start=855
  _globals['_GETRUNSTATUSRESPONSE']._serialized_end=1032
  _globals['_GETRUNSTATUSRESPONSE_RUNSTATUSDICTENTRY']._serialized_start=957
  _globals['_GETRUNSTATUSRESPONSE_RUNSTATUSDICTENTRY']._serialized_end=1032
  _globals['_GETFEDERATIONOPTIONSREQUEST']._serialized_start=1034
  _globals['_GETFEDERATIONOPTIONSREQUEST']._serialized_end=1079
  _globals['_GETFEDERATIONOPTIONSRESPONSE']._serialized_start=1081
  _globals['_GETFEDERATIONOPTIONSRESPONSE']._serialized_end=1165
# @@protoc_insertion_point(module_scope)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower account auth plugins."""


from flwr.common.constant import AuthnType

from .auth_plugin import CliAuthPlugin, LoginError
from .noop_auth_plugin import NoOpCliAuthPlugin
from .oidc_cli_plugin import OidcCliPlugin


def get_cli_plugin_class(authn_type: str) -> type[CliAuthPlugin]:
    """Return all CLI authentication plugins."""
    if authn_type == AuthnType.NOOP:
        return NoOpCliAuthPlugin
    if authn_type == AuthnType.OIDC:
        return OidcCliPlugin
    raise ValueError(f"Unsupported authentication type: {authn_type}")


__all__ = [
    "CliAuthPlugin",
    "LoginError",
    "NoOpCliAuthPlugin",
    "OidcCliPlugin",
    "get_cli_plugin_class",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import flwr.proto.node_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Fab(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class VerificationsEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: builtins.str = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    HASH_STR_FIELD_NUMBER: builtins.int
    CONTENT_FIELD_NUMBER: builtins.int
    VERIFICATIONS_FIELD_NUMBER: builtins.int
    hash_str: builtins.str
    """This field is the hash of the data field. It is used to identify the data.
    The hash is calculated using the SHA-256 algorithm and is represented as a
    hex string (sha256hex).
    """
    content: builtins.bytes
    """This field contains the fab file contents a one bytes blob."""
    @property
    def verifications(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """Verifications."""

    def __init__(
        self,
        *,
        hash_str: builtins.str = ...,
        content: builtins.bytes = ...,
        verifications: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["content", b"content", "hash_str", b"hash_str", "verifications", b"verifications"]) -> None: ...

global___Fab = Fab

@typing.final
class GetFabRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    HASH_STR_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    hash_str: builtins.str
    run_id: builtins.int
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        hash_str: builtins.str = ...,
        run_id: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["hash_str", b"hash_str", "node", b"node", "run_id", b"run_id"]) -> None: ...

global___GetFabRequest = GetFabRequest

@typing.final
class GetFabResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FAB_FIELD_NUMBER: builtins.int
    @property
    def fab(self) -> global___Fab: ...
    def __init__(
        self,
        *,
        fab: global___Fab | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["fab", b"fab"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["fab", b"fab"]) -> None: ...

global___GetFabResponse = GetFabResponse



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `federation list` command."""


import io
from pathlib import Path
from typing import Annotated, cast

import typer
from rich.console import Console
from rich.table import Table
from rich.text import Text

from flwr.cli.config_utils import (
    exit_if_no_address,
    load_and_validate,
    process_loaded_project_config,
    validate_federation_in_project_config,
)
from flwr.common.constant import FAB_CONFIG_FILE, CliOutputFormat
from flwr.common.logger import print_json_error, redirect_output, restore_output
from flwr.proto.control_pb2 import (  # pylint: disable=E0611
    ListFederationsRequest,
    ListFederationsResponse,
)
from flwr.proto.control_pb2_grpc import ControlStub
from flwr.proto.federation_pb2 import Federation  # pylint: disable=E0611

from ..utils import flwr_cli_grpc_exc_handler, init_channel, load_cli_auth_plugin


def ls(  # pylint: disable=R0914, R0913, R0917
    ctx: typer.Context,
    app: Annotated[
        Path,
        typer.Argument(help="Path of the Flower project"),
    ] = Path("."),
    federation: Annotated[
        str | None,
        typer.Argument(help="Name of the federation"),
    ] = None,
    output_format: Annotated[
        str,
        typer.Option(
            "--format",
            case_sensitive=False,
            help="Format output using 'default' view or 'json'",
        ),
    ] = CliOutputFormat.DEFAULT,
) -> None:
    """List available federations."""
    # Resolve command used (list or ls)
    command_name = cast(str, ctx.command.name) if ctx.command else "ls"

    suppress_output = output_format == CliOutputFormat.JSON
    captured_output = io.StringIO()
    try:
        if suppress_output:
            redirect_output(captured_output)
        typer.secho("Loading project configuration... ", fg=typer.colors.BLUE)

        pyproject_path = app / FAB_CONFIG_FILE if app else None
        config, errors, warnings = load_and_validate(pyproject_path, check_module=False)
        config = process_loaded_project_config(config, errors, warnings)
        federation, federation_config = validate_federation_in_project_config(
            federation, config
        )
        exit_if_no_address(federation_config, f"federation {command_name}")
        channel = None
        try:
            auth_plugin = load_cli_auth_plugin(app, federation, federation_config)
            channel = init_channel(app, federation_config, auth_plugin)
            stub = ControlStub(channel)
            typer.echo("📄 Listing federations...")
            federations = _list_federations(stub)
            restore_output()
            if output_format == CliOutputFormat.JSON:
                Console().print_json(data=_to_json(federations))
            else:
                Console().print(_to_table(federations))
        finally:
            if channel:
                channel.close()
    except (typer.Exit, Exception) as err:  # pylint: disable=broad-except
        if suppress_output:
            restore_output()
            e_message = captured_output.getvalue()
            print_json_error(e_message, err)
        else:
            typer.secho(
                f"{err}",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
    finally:
        if suppress_output:
            restore_output()
        captured_output.close()


def _list_federations(stub: ControlStub) -> list[Federation]:
    """List all federations."""
    with flwr_cli_grpc_exc_handler():
        res: ListFederationsResponse = stub.ListFederations(ListFederationsRequest())

    return list(res.federations)


def _to_table(federations: list[Federation]) -> Table:
    """Format the provided federations list to a rich Table."""
    table = Table(header_style="bold cyan", show_lines=True)

    # Add columns
    table.add_column(
        Text("Federation", justify="center"), style="bright_black", no_wrap=True
    )

    for federation in federations:
        table.add_row(federation.name)

    return table


def _to_json(federations: list[Federation]) -> list[dict[str, str]]:
    """Format the provided federations list to JSON serializable format."""
    return [{"name": federation.name} for federation in federations]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2025 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import flwr.proto.fab_pb2
import flwr.proto.message_pb2
import flwr.proto.run_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class ListAppsToLaunchRequest(google.protobuf.message.Message):
    """These messages are used by both ServerAppIo and ClientAppIo services

    ListAppsToLaunch messages
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___ListAppsToLaunchRequest = ListAppsToLaunchRequest

@typing.final
class ListAppsToLaunchResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_IDS_FIELD_NUMBER: builtins.int
    @property
    def run_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """List of run IDs of the apps to launch"""

    def __init__(
        self,
        *,
        run_ids: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["run_ids", b"run_ids"]) -> None: ...

global___ListAppsToLaunchResponse = ListAppsToLaunchResponse

@typing.final
class RequestTokenRequest(google.protobuf.message.Message):
    """RequestToken messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_ID_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["run_id", b"run_id"]) -> None: ...

global___RequestTokenRequest = RequestTokenRequest

@typing.final
class RequestTokenResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    token: builtins.str
    def __init__(
        self,
        *,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["token", b"token"]) -> None: ...

global___RequestTokenResponse = RequestTokenResponse

@typing.final
class PushAppMessagesRequest(google.protobuf.message.Message):
    """PushAppMessages messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    MESSAGES_LIST_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    MESSAGE_OBJECT_TREES_FIELD_NUMBER: builtins.int
    token: builtins.str
    run_id: builtins.int
    @property
    def messages_list(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[flwr.proto.message_pb2.Message]: ...
    @property
    def message_object_trees(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[flwr.proto.message_pb2.ObjectTree]: ...
    def __init__(
        self,
        *,
        token: builtins.str = ...,
        messages_list: collections.abc.Iterable[flwr.proto.message_pb2.Message] | None = ...,
        run_id: builtins.int = ...,
        message_object_trees: collections.abc.Iterable[flwr.proto.message_pb2.ObjectTree] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["message_object_trees", b"message_object_trees", "messages_list", b"messages_list", "run_id", b"run_id", "token", b"token"]) -> None: ...

global___PushAppMessagesRequest = PushAppMessagesRequest

@typing.final
class PushAppMessagesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGE_IDS_FIELD_NUMBER: builtins.int
    OBJECTS_TO_PUSH_FIELD_NUMBER: builtins.int
    @property
    def message_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    @property
    def objects_to_push(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        message_ids: collections.abc.Iterable[builtins.str] | None = ...,
        objects_to_push: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["message_ids", b"message_ids", "objects_to_push", b"objects_to_push"]) -> None: ...

global___PushAppMessagesResponse = PushAppMessagesResponse

@typing.final
class PullAppMessagesRequest(google.protobuf.message.Message):
    """PullAppMessages messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    MESSAGE_IDS_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    token: builtins.str
    run_id: builtins.int
    @property
    def message_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        token: builtins.str = ...,
        message_ids: collections.abc.Iterable[builtins.str] | None = ...,
        run_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["message_ids", b"message_ids", "run_id", b"run_id", "token", b"token"]) -> None: ...

global___PullAppMessagesRequest = PullAppMessagesRequest

@typing.final
class PullAppMessagesResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    MESSAGES_LIST_FIELD_NUMBER: builtins.int
    MESSAGE_OBJECT_TREES_FIELD_NUMBER: builtins.int
    @property
    def messages_list(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[flwr.proto.message_pb2.Message]: ...
    @property
    def message_object_trees(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[flwr.proto.message_pb2.ObjectTree]: ...
    def __init__(
        self,
        *,
        messages_list: collections.abc.Iterable[flwr.proto.message_pb2.Message] | None = ...,
        message_object_trees: collections.abc.Iterable[flwr.proto.message_pb2.ObjectTree] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["message_object_trees", b"message_object_trees", "messages_list", b"messages_list"]) -> None: ...

global___PullAppMessagesResponse = PullAppMessagesResponse

@typing.final
class PullAppInputsRequest(google.protobuf.message.Message):
    """PullAppInputs messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    token: builtins.str
    def __init__(
        self,
        *,
        token: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["token", b"token"]) -> None: ...

global___PullAppInputsRequest = PullAppInputsRequest

@typing.final
class PullAppInputsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CONTEXT_FIELD_NUMBER: builtins.int
    RUN_FIELD_NUMBER: builtins.int
    FAB_FIELD_NUMBER: builtins.int
    @property
    def context(self) -> flwr.proto.message_pb2.Context: ...
    @property
    def run(self) -> flwr.proto.run_pb2.Run: ...
    @property
    def fab(self) -> flwr.proto.fab_pb2.Fab: ...
    def __init__(
        self,
        *,
        context: flwr.proto.message_pb2.Context | None = ...,
        run: flwr.proto.run_pb2.Run | None = ...,
        fab: flwr.proto.fab_pb2.Fab | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context", "fab", b"fab", "run", b"run"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "fab", b"fab", "run", b"run"]) -> None: ...

global___PullAppInputsResponse = PullAppInputsResponse

@typing.final
class PushAppOutputsRequest(google.protobuf.message.Message):
    """PushAppInputs messages"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    CONTEXT_FIELD_NUMBER: builtins.int
    token: builtins.str
    run_id: builtins.int
    @property
    def context(self) -> flwr.proto.message_pb2.Context: ...
    def __init__(
        self,
        *,
        token: builtins.str = ...,
        run_id: builtins.int = ...,
        context: flwr.proto.message_pb2.Context | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["context", b"context"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["context", b"context", "run_id", b"run_id", "token", b"token"]) -> None: ...

global___PushAppOutputsRequest = PushAppOutputsRequest

@typing.final
class PushAppOutputsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___PushAppOutputsResponse = PushAppOutputsResponse



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Adaptive Federated Optimization using Adam (FedAdam) strategy.

[Reddi et al., 2020]

Paper: arxiv.org/abs/2003.00295
"""


from collections.abc import Callable, Iterable

import numpy as np

from flwr.common import Array, ArrayRecord, Message, MetricRecord, RecordDict

from ..exception import AggregationError
from .fedopt import FedOpt


# pylint: disable=line-too-long
class FedAdam(FedOpt):
    """FedAdam - Adaptive Federated Optimization using Adam.

    Implementation based on https://arxiv.org/abs/2003.00295v5

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for both ArrayRecords and MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
         Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    eta : float, optional
        Server-side learning rate. Defaults to 1e-1.
    eta_l : float, optional
        Client-side learning rate. Defaults to 1e-1.
    beta_1 : float, optional
        Momentum parameter. Defaults to 0.9.
    beta_2 : float, optional
        Second moment parameter. Defaults to 0.99.
    tau : float, optional
        Controls the algorithm's degree of adaptability. Defaults to 1e-3.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    def __init__(
        self,
        *,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_train_nodes: int = 2,
        min_evaluate_nodes: int = 2,
        min_available_nodes: int = 2,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        eta: float = 1e-1,
        eta_l: float = 1e-1,
        beta_1: float = 0.9,
        beta_2: float = 0.99,
        tau: float = 1e-3,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=min_train_nodes,
            min_evaluate_nodes=min_evaluate_nodes,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
            eta=eta,
            eta_l=eta_l,
            beta_1=beta_1,
            beta_2=beta_2,
            tau=tau,
        )

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        aggregated_arrayrecord, aggregated_metrics = super().aggregate_train(
            server_round, replies
        )

        if aggregated_arrayrecord is None:
            return aggregated_arrayrecord, aggregated_metrics

        if self.current_arrays is None:
            reason = (
                "Current arrays not set. Ensure that `configure_train` has been "
                "called before aggregation."
            )
            raise AggregationError(reason=reason)

        # Compute intermediate variables
        delta_t, m_t, aggregated_ndarrays = self._compute_deltat_and_mt(
            aggregated_arrayrecord
        )

        # v_t
        if not self.v_t:
            self.v_t = {k: np.zeros_like(v) for k, v in aggregated_ndarrays.items()}
        self.v_t = {
            k: self.beta_2 * v + (1 - self.beta_2) * (delta_t[k] ** 2)
            for k, v in self.v_t.items()
        }

        # Compute the bias-corrected learning rate, `eta_norm` for improving convergence
        # in the early rounds of FL training. This `eta_norm` is `\alpha_t` in Kingma &
        # Ba, 2014 (http://arxiv.org/abs/1412.6980) "Adam: A Method for Stochastic
        # Optimization" in the formula line right before Section 2.1.
        eta_norm = (
            self.eta
            * np.sqrt(1 - np.power(self.beta_2, server_round + 1.0))
            / (1 - np.power(self.beta_1, server_round + 1.0))
        )

        new_arrays = {
            k: x + eta_norm * m_t[k] / (np.sqrt(self.v_t[k]) + self.tau)
            for k, x in self.current_arrays.items()
        }

        return (
            ArrayRecord({k: Array(v) for k, v in new_arrays.items()}),
            aggregated_metrics,
        )



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Machine Learning with Adversaries: Byzantine Tolerant Gradient Descent.

[Blanchard et al., 2017].

Paper: proceedings.neurips.cc/paper/2017/file/f4b9ec30ad9f68f89b29639786cb62ef-Paper.pdf
"""


from collections.abc import Callable, Iterable
from logging import INFO
from typing import cast

import numpy as np

from flwr.common import ArrayRecord, Message, MetricRecord, NDArray, RecordDict, log

from .fedavg import FedAvg
from .strategy_utils import aggregate_arrayrecords


# pylint: disable=too-many-instance-attributes
class MultiKrum(FedAvg):
    """MultiKrum [Blanchard et al., 2017] strategy.

    Implementation based on https://arxiv.org/abs/1703.02757

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    num_malicious_nodes : int (default: 0)
        Number of malicious nodes in the system. Defaults to 0.
    num_nodes_to_select : int (default: 1)
        Number of nodes to select before averaging.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for both ArrayRecords and MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.

    Notes
    -----
    MultiKrum is a generalization of Krum. If `num_nodes_to_select` is set to 1,
    MultiKrum will reduce to classical Krum.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_train_nodes: int = 2,
        min_evaluate_nodes: int = 2,
        min_available_nodes: int = 2,
        num_malicious_nodes: int = 0,
        num_nodes_to_select: int = 1,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=min_train_nodes,
            min_evaluate_nodes=min_evaluate_nodes,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
        )
        self.num_malicious_nodes = num_malicious_nodes
        self.num_nodes_to_select = num_nodes_to_select

    def summary(self) -> None:
        """Log summary configuration of the strategy."""
        log(INFO, "\t├──> MultiKrum settings:")
        log(INFO, "\t│\t├── Number of malicious nodes: %d", self.num_malicious_nodes)
        log(INFO, "\t│\t└── Number of nodes to select: %d", self.num_nodes_to_select)
        super().summary()

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        valid_replies, _ = self._check_and_log_replies(replies, is_train=True)

        arrays, metrics = None, None
        if valid_replies:
            reply_contents = [msg.content for msg in valid_replies]

            # Krum or MultiKrum selection
            replies_to_aggregate = select_multikrum(
                reply_contents,
                num_malicious_nodes=self.num_malicious_nodes,
                num_nodes_to_select=self.num_nodes_to_select,
            )

            # Aggregate ArrayRecords
            arrays = aggregate_arrayrecords(
                replies_to_aggregate,
                self.weighted_by_key,
            )

            # Aggregate MetricRecords
            metrics = self.train_metrics_aggr_fn(
                replies_to_aggregate,
                self.weighted_by_key,
            )
        return arrays, metrics


def compute_distances(records: list[ArrayRecord]) -> NDArray:
    """Compute squared L2 distances between ArrayRecords.

    Parameters
    ----------
    records : list[ArrayRecord]
        A list of ArrayRecords (arrays received in replies)

    Returns
    -------
    NDArray
        A 2D array representing the distance matrix of squared L2 distances
        between input ArrayRecords
    """
    # Formula: ||x - y||^2 = ||x||^2 + ||y||^2 - 2 * x.y
    # Flatten records and stack them into a matrix
    flat_w = np.stack(
        [np.concatenate(rec.to_numpy_ndarrays(), axis=None).ravel() for rec in records],
        axis=0,
    )  # shape: (n, d) with n number of records and d the dimension of model

    # Compute squared norms of each vector
    norms: NDArray = np.square(flat_w).sum(axis=1)  # shape (n,)

    # Use broadcasting to compute pairwise distances
    distance_matrix: NDArray = norms[:, None] + norms[None, :] - 2 * flat_w @ flat_w.T
    return distance_matrix


def select_multikrum(
    contents: list[RecordDict],
    num_malicious_nodes: int,
    num_nodes_to_select: int,
) -> list[RecordDict]:
    """Select the set of RecordDicts to aggregate using the Krum or MultiKrum algorithm.

    For each node, computes the sum of squared L2 distances to its n-f-2 closest
    parameter vectors, where n is the number of nodes and f is the number of
    malicious nodes. The node(s) with the lowest score(s) are selected for
    aggregation.

    Parameters
    ----------
    contents : list[RecordDict]
        List of contents from reply messages, where each content is a RecordDict
        containing an ArrayRecord of model parameters from a node (client).
    num_malicious_nodes : int
        Number of malicious nodes in the system.
    num_nodes_to_select : int
        Number of client updates to select.
        - If 1, the algorithm reduces to classical Krum (selecting a single update).
        - If >1, Multi-Krum is applied (selecting multiple updates).

    Returns
    -------
    list[RecordDict]
        Selected contents following the Krum or Multi-Krum algorithm.

    Notes
    -----
    If `num_nodes_to_select` is set to 1, Multi-Krum reduces to classical Krum
    and only a single RecordDict is selected.
    """
    # Construct list of ArrayRecord objects from replies
    record_key = list(contents[0].array_records.keys())[0]
    # Recall aggregate_train first ensures replies only contain one ArrayRecord
    array_records = [cast(ArrayRecord, reply[record_key]) for reply in contents]
    distance_matrix = compute_distances(array_records)

    # For each node, take the n-f-2 closest parameters vectors
    num_closest = max(1, len(array_records) - num_malicious_nodes - 2)
    closest_indices = []
    for distance in distance_matrix:
        closest_indices.append(
            np.argsort(distance)[1 : num_closest + 1].tolist()  # noqa: E203
        )

    # Compute the score for each node, that is the sum of the distances
    # of the n-f-2 closest parameters vectors
    scores = [
        np.sum(distance_matrix[i, closest_indices[i]])
        for i in range(len(distance_matrix))
    ]

    # Choose the num_nodes_to_select lowest-scoring nodes (MultiKrum)
    # and return their updates
    best_indices = np.argsort(scores)[:num_nodes_to_select]
    return [contents[i] for i in best_indices]



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Machine Learning with Adversaries: Byzantine Tolerant Gradient Descent.

[Blanchard et al., 2017].

Paper: proceedings.neurips.cc/paper/2017/file/f4b9ec30ad9f68f89b29639786cb62ef-Paper.pdf
"""


from collections.abc import Callable
from logging import INFO

from flwr.common import MetricRecord, RecordDict, log

from .multikrum import MultiKrum


# pylint: disable=too-many-instance-attributes
class Krum(MultiKrum):
    """Krum [Blanchard et al., 2017] strategy.

    Implementation based on https://arxiv.org/abs/1703.02757

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    num_malicious_nodes : int (default: 0)
        Number of malicious nodes in the system. Defaults to 0.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    """

    # pylint: disable=too-many-arguments,too-many-positional-arguments
    def __init__(
        self,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_train_nodes: int = 2,
        min_evaluate_nodes: int = 2,
        min_available_nodes: int = 2,
        num_malicious_nodes: int = 0,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=min_train_nodes,
            min_evaluate_nodes=min_evaluate_nodes,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            num_malicious_nodes=num_malicious_nodes,
            num_nodes_to_select=1,  # Krum selects 1 node
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
        )

    def summary(self) -> None:
        """Log summary configuration of the strategy."""
        log(INFO, "\t├──> Krum settings:")
        log(INFO, "\t│\t└── Number of malicious nodes: %d", self.num_malicious_nodes)
        super().summary()



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _Code:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _CodeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Code.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    OK: _Code.ValueType  # 0
    GET_PROPERTIES_NOT_IMPLEMENTED: _Code.ValueType  # 1
    GET_PARAMETERS_NOT_IMPLEMENTED: _Code.ValueType  # 2
    FIT_NOT_IMPLEMENTED: _Code.ValueType  # 3
    EVALUATE_NOT_IMPLEMENTED: _Code.ValueType  # 4

class Code(_Code, metaclass=_CodeEnumTypeWrapper): ...

OK: Code.ValueType  # 0
GET_PROPERTIES_NOT_IMPLEMENTED: Code.ValueType  # 1
GET_PARAMETERS_NOT_IMPLEMENTED: Code.ValueType  # 2
FIT_NOT_IMPLEMENTED: Code.ValueType  # 3
EVALUATE_NOT_IMPLEMENTED: Code.ValueType  # 4
global___Code = Code

class _Reason:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _ReasonEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Reason.ValueType], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNKNOWN: _Reason.ValueType  # 0
    RECONNECT: _Reason.ValueType  # 1
    POWER_DISCONNECTED: _Reason.ValueType  # 2
    WIFI_UNAVAILABLE: _Reason.ValueType  # 3
    ACK: _Reason.ValueType  # 4

class Reason(_Reason, metaclass=_ReasonEnumTypeWrapper): ...

UNKNOWN: Reason.ValueType  # 0
RECONNECT: Reason.ValueType  # 1
POWER_DISCONNECTED: Reason.ValueType  # 2
WIFI_UNAVAILABLE: Reason.ValueType  # 3
ACK: Reason.ValueType  # 4
global___Reason = Reason

@typing.final
class Status(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    CODE_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    code: global___Code.ValueType
    message: builtins.str
    def __init__(
        self,
        *,
        code: global___Code.ValueType = ...,
        message: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["code", b"code", "message", b"message"]) -> None: ...

global___Status = Status

@typing.final
class Parameters(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TENSORS_FIELD_NUMBER: builtins.int
    TENSOR_TYPE_FIELD_NUMBER: builtins.int
    tensor_type: builtins.str
    @property
    def tensors(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        tensors: collections.abc.Iterable[builtins.bytes] | None = ...,
        tensor_type: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["tensor_type", b"tensor_type", "tensors", b"tensors"]) -> None: ...

global___Parameters = Parameters

@typing.final
class ServerMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class ReconnectIns(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        SECONDS_FIELD_NUMBER: builtins.int
        seconds: builtins.int
        def __init__(
            self,
            *,
            seconds: builtins.int = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["seconds", b"seconds"]) -> None: ...

    @typing.final
    class GetPropertiesIns(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ConfigEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        CONFIG_FIELD_NUMBER: builtins.int
        @property
        def config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            config: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["config", b"config"]) -> None: ...

    @typing.final
    class GetParametersIns(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ConfigEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        CONFIG_FIELD_NUMBER: builtins.int
        @property
        def config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            config: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["config", b"config"]) -> None: ...

    @typing.final
    class FitIns(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ConfigEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        PARAMETERS_FIELD_NUMBER: builtins.int
        CONFIG_FIELD_NUMBER: builtins.int
        @property
        def parameters(self) -> global___Parameters: ...
        @property
        def config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            parameters: global___Parameters | None = ...,
            config: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["parameters", b"parameters"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["config", b"config", "parameters", b"parameters"]) -> None: ...

    @typing.final
    class EvaluateIns(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class ConfigEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        PARAMETERS_FIELD_NUMBER: builtins.int
        CONFIG_FIELD_NUMBER: builtins.int
        @property
        def parameters(self) -> global___Parameters: ...
        @property
        def config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            parameters: global___Parameters | None = ...,
            config: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["parameters", b"parameters"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["config", b"config", "parameters", b"parameters"]) -> None: ...

    RECONNECT_INS_FIELD_NUMBER: builtins.int
    GET_PROPERTIES_INS_FIELD_NUMBER: builtins.int
    GET_PARAMETERS_INS_FIELD_NUMBER: builtins.int
    FIT_INS_FIELD_NUMBER: builtins.int
    EVALUATE_INS_FIELD_NUMBER: builtins.int
    @property
    def reconnect_ins(self) -> global___ServerMessage.ReconnectIns: ...
    @property
    def get_properties_ins(self) -> global___ServerMessage.GetPropertiesIns: ...
    @property
    def get_parameters_ins(self) -> global___ServerMessage.GetParametersIns: ...
    @property
    def fit_ins(self) -> global___ServerMessage.FitIns: ...
    @property
    def evaluate_ins(self) -> global___ServerMessage.EvaluateIns: ...
    def __init__(
        self,
        *,
        reconnect_ins: global___ServerMessage.ReconnectIns | None = ...,
        get_properties_ins: global___ServerMessage.GetPropertiesIns | None = ...,
        get_parameters_ins: global___ServerMessage.GetParametersIns | None = ...,
        fit_ins: global___ServerMessage.FitIns | None = ...,
        evaluate_ins: global___ServerMessage.EvaluateIns | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["evaluate_ins", b"evaluate_ins", "fit_ins", b"fit_ins", "get_parameters_ins", b"get_parameters_ins", "get_properties_ins", b"get_properties_ins", "msg", b"msg", "reconnect_ins", b"reconnect_ins"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["evaluate_ins", b"evaluate_ins", "fit_ins", b"fit_ins", "get_parameters_ins", b"get_parameters_ins", "get_properties_ins", b"get_properties_ins", "msg", b"msg", "reconnect_ins", b"reconnect_ins"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["msg", b"msg"]) -> typing.Literal["reconnect_ins", "get_properties_ins", "get_parameters_ins", "fit_ins", "evaluate_ins"] | None: ...

global___ServerMessage = ServerMessage

@typing.final
class ClientMessage(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class DisconnectRes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        REASON_FIELD_NUMBER: builtins.int
        reason: global___Reason.ValueType
        def __init__(
            self,
            *,
            reason: global___Reason.ValueType = ...,
        ) -> None: ...
        def ClearField(self, field_name: typing.Literal["reason", b"reason"]) -> None: ...

    @typing.final
    class GetPropertiesRes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class PropertiesEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        STATUS_FIELD_NUMBER: builtins.int
        PROPERTIES_FIELD_NUMBER: builtins.int
        @property
        def status(self) -> global___Status: ...
        @property
        def properties(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            status: global___Status | None = ...,
            properties: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["status", b"status"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["properties", b"properties", "status", b"status"]) -> None: ...

    @typing.final
    class GetParametersRes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        STATUS_FIELD_NUMBER: builtins.int
        PARAMETERS_FIELD_NUMBER: builtins.int
        @property
        def status(self) -> global___Status: ...
        @property
        def parameters(self) -> global___Parameters: ...
        def __init__(
            self,
            *,
            status: global___Status | None = ...,
            parameters: global___Parameters | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["parameters", b"parameters", "status", b"status"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["parameters", b"parameters", "status", b"status"]) -> None: ...

    @typing.final
    class FitRes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class MetricsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        STATUS_FIELD_NUMBER: builtins.int
        PARAMETERS_FIELD_NUMBER: builtins.int
        NUM_EXAMPLES_FIELD_NUMBER: builtins.int
        METRICS_FIELD_NUMBER: builtins.int
        num_examples: builtins.int
        @property
        def status(self) -> global___Status: ...
        @property
        def parameters(self) -> global___Parameters: ...
        @property
        def metrics(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            status: global___Status | None = ...,
            parameters: global___Parameters | None = ...,
            num_examples: builtins.int = ...,
            metrics: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["parameters", b"parameters", "status", b"status"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["metrics", b"metrics", "num_examples", b"num_examples", "parameters", b"parameters", "status", b"status"]) -> None: ...

    @typing.final
    class EvaluateRes(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        @typing.final
        class MetricsEntry(google.protobuf.message.Message):
            DESCRIPTOR: google.protobuf.descriptor.Descriptor

            KEY_FIELD_NUMBER: builtins.int
            VALUE_FIELD_NUMBER: builtins.int
            key: builtins.str
            @property
            def value(self) -> global___Scalar: ...
            def __init__(
                self,
                *,
                key: builtins.str = ...,
                value: global___Scalar | None = ...,
            ) -> None: ...
            def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
            def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

        STATUS_FIELD_NUMBER: builtins.int
        LOSS_FIELD_NUMBER: builtins.int
        NUM_EXAMPLES_FIELD_NUMBER: builtins.int
        METRICS_FIELD_NUMBER: builtins.int
        loss: builtins.float
        num_examples: builtins.int
        @property
        def status(self) -> global___Status: ...
        @property
        def metrics(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Scalar]: ...
        def __init__(
            self,
            *,
            status: global___Status | None = ...,
            loss: builtins.float = ...,
            num_examples: builtins.int = ...,
            metrics: collections.abc.Mapping[builtins.str, global___Scalar] | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["status", b"status"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["loss", b"loss", "metrics", b"metrics", "num_examples", b"num_examples", "status", b"status"]) -> None: ...

    DISCONNECT_RES_FIELD_NUMBER: builtins.int
    GET_PROPERTIES_RES_FIELD_NUMBER: builtins.int
    GET_PARAMETERS_RES_FIELD_NUMBER: builtins.int
    FIT_RES_FIELD_NUMBER: builtins.int
    EVALUATE_RES_FIELD_NUMBER: builtins.int
    @property
    def disconnect_res(self) -> global___ClientMessage.DisconnectRes: ...
    @property
    def get_properties_res(self) -> global___ClientMessage.GetPropertiesRes: ...
    @property
    def get_parameters_res(self) -> global___ClientMessage.GetParametersRes: ...
    @property
    def fit_res(self) -> global___ClientMessage.FitRes: ...
    @property
    def evaluate_res(self) -> global___ClientMessage.EvaluateRes: ...
    def __init__(
        self,
        *,
        disconnect_res: global___ClientMessage.DisconnectRes | None = ...,
        get_properties_res: global___ClientMessage.GetPropertiesRes | None = ...,
        get_parameters_res: global___ClientMessage.GetParametersRes | None = ...,
        fit_res: global___ClientMessage.FitRes | None = ...,
        evaluate_res: global___ClientMessage.EvaluateRes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["disconnect_res", b"disconnect_res", "evaluate_res", b"evaluate_res", "fit_res", b"fit_res", "get_parameters_res", b"get_parameters_res", "get_properties_res", b"get_properties_res", "msg", b"msg"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["disconnect_res", b"disconnect_res", "evaluate_res", b"evaluate_res", "fit_res", b"fit_res", "get_parameters_res", b"get_parameters_res", "get_properties_res", b"get_properties_res", "msg", b"msg"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["msg", b"msg"]) -> typing.Literal["disconnect_res", "get_properties_res", "get_parameters_res", "fit_res", "evaluate_res"] | None: ...

global___ClientMessage = ClientMessage

@typing.final
class Scalar(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DOUBLE_FIELD_NUMBER: builtins.int
    UINT64_FIELD_NUMBER: builtins.int
    SINT64_FIELD_NUMBER: builtins.int
    BOOL_FIELD_NUMBER: builtins.int
    STRING_FIELD_NUMBER: builtins.int
    BYTES_FIELD_NUMBER: builtins.int
    double: builtins.float
    uint64: builtins.int
    """float float = 2;
    int32 int32 = 3;
    int64 int64 = 4;
    uint32 uint32 = 5;
    """
    sint64: builtins.int
    """sint32 sint32 = 7;"""
    bool: builtins.bool
    """fixed32 fixed32 = 9;
    fixed64 fixed64 = 10;
    sfixed32 sfixed32 = 11;
    sfixed64 sfixed64 = 12;
    """
    string: builtins.str
    bytes: builtins.bytes
    def __init__(
        self,
        *,
        double: builtins.float = ...,
        uint64: builtins.int = ...,
        sint64: builtins.int = ...,
        bool: builtins.bool = ...,
        string: builtins.str = ...,
        bytes: builtins.bytes = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bool", b"bool", "bytes", b"bytes", "double", b"double", "scalar", b"scalar", "sint64", b"sint64", "string", b"string", "uint64", b"uint64"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bool", b"bool", "bytes", b"bytes", "double", b"double", "scalar", b"scalar", "sint64", b"sint64", "string", b"string", "uint64", b"uint64"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["scalar", b"scalar"]) -> typing.Literal["double", "uint64", "sint64", "bool", "string", "bytes"] | None: ...

global___Scalar = Scalar



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import flwr.proto.node_pb2
import flwr.proto.recorddict_pb2
import flwr.proto.transport_pb2
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class Run(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class OverrideConfigEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> flwr.proto.transport_pb2.Scalar: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: flwr.proto.transport_pb2.Scalar | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    RUN_ID_FIELD_NUMBER: builtins.int
    FAB_ID_FIELD_NUMBER: builtins.int
    FAB_VERSION_FIELD_NUMBER: builtins.int
    OVERRIDE_CONFIG_FIELD_NUMBER: builtins.int
    FAB_HASH_FIELD_NUMBER: builtins.int
    PENDING_AT_FIELD_NUMBER: builtins.int
    STARTING_AT_FIELD_NUMBER: builtins.int
    RUNNING_AT_FIELD_NUMBER: builtins.int
    FINISHED_AT_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    FLWR_AID_FIELD_NUMBER: builtins.int
    FEDERATION_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    fab_id: builtins.str
    fab_version: builtins.str
    fab_hash: builtins.str
    pending_at: builtins.str
    starting_at: builtins.str
    running_at: builtins.str
    finished_at: builtins.str
    flwr_aid: builtins.str
    federation: builtins.str
    @property
    def override_config(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, flwr.proto.transport_pb2.Scalar]: ...
    @property
    def status(self) -> global___RunStatus: ...
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
        fab_id: builtins.str = ...,
        fab_version: builtins.str = ...,
        override_config: collections.abc.Mapping[builtins.str, flwr.proto.transport_pb2.Scalar] | None = ...,
        fab_hash: builtins.str = ...,
        pending_at: builtins.str = ...,
        starting_at: builtins.str = ...,
        running_at: builtins.str = ...,
        finished_at: builtins.str = ...,
        status: global___RunStatus | None = ...,
        flwr_aid: builtins.str = ...,
        federation: builtins.str = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["status", b"status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["fab_hash", b"fab_hash", "fab_id", b"fab_id", "fab_version", b"fab_version", "federation", b"federation", "finished_at", b"finished_at", "flwr_aid", b"flwr_aid", "override_config", b"override_config", "pending_at", b"pending_at", "run_id", b"run_id", "running_at", b"running_at", "starting_at", b"starting_at", "status", b"status"]) -> None: ...

global___Run = Run

@typing.final
class RunStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATUS_FIELD_NUMBER: builtins.int
    SUB_STATUS_FIELD_NUMBER: builtins.int
    DETAILS_FIELD_NUMBER: builtins.int
    status: builtins.str
    """"starting", "running", "finished" """
    sub_status: builtins.str
    """"completed", "failed", "stopped" or "" (non-finished)"""
    details: builtins.str
    """failure details"""
    def __init__(
        self,
        *,
        status: builtins.str = ...,
        sub_status: builtins.str = ...,
        details: builtins.str = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["details", b"details", "status", b"status", "sub_status", b"sub_status"]) -> None: ...

global___RunStatus = RunStatus

@typing.final
class GetRunRequest(google.protobuf.message.Message):
    """GetRun"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    RUN_ID_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        run_id: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["node", b"node", "run_id", b"run_id"]) -> None: ...

global___GetRunRequest = GetRunRequest

@typing.final
class GetRunResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_FIELD_NUMBER: builtins.int
    @property
    def run(self) -> global___Run: ...
    def __init__(
        self,
        *,
        run: global___Run | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["run", b"run"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["run", b"run"]) -> None: ...

global___GetRunResponse = GetRunResponse

@typing.final
class UpdateRunStatusRequest(google.protobuf.message.Message):
    """UpdateRunStatus"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_ID_FIELD_NUMBER: builtins.int
    RUN_STATUS_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    @property
    def run_status(self) -> global___RunStatus: ...
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
        run_status: global___RunStatus | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["run_status", b"run_status"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["run_id", b"run_id", "run_status", b"run_status"]) -> None: ...

global___UpdateRunStatusRequest = UpdateRunStatusRequest

@typing.final
class UpdateRunStatusResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    def __init__(
        self,
    ) -> None: ...

global___UpdateRunStatusResponse = UpdateRunStatusResponse

@typing.final
class GetRunStatusRequest(google.protobuf.message.Message):
    """GetRunStatus"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NODE_FIELD_NUMBER: builtins.int
    RUN_IDS_FIELD_NUMBER: builtins.int
    @property
    def node(self) -> flwr.proto.node_pb2.Node: ...
    @property
    def run_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        node: flwr.proto.node_pb2.Node | None = ...,
        run_ids: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["node", b"node"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["node", b"node", "run_ids", b"run_ids"]) -> None: ...

global___GetRunStatusRequest = GetRunStatusRequest

@typing.final
class GetRunStatusResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class RunStatusDictEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.int
        @property
        def value(self) -> global___RunStatus: ...
        def __init__(
            self,
            *,
            key: builtins.int = ...,
            value: global___RunStatus | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    RUN_STATUS_DICT_FIELD_NUMBER: builtins.int
    @property
    def run_status_dict(self) -> google.protobuf.internal.containers.MessageMap[builtins.int, global___RunStatus]: ...
    def __init__(
        self,
        *,
        run_status_dict: collections.abc.Mapping[builtins.int, global___RunStatus] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["run_status_dict", b"run_status_dict"]) -> None: ...

global___GetRunStatusResponse = GetRunStatusResponse

@typing.final
class GetFederationOptionsRequest(google.protobuf.message.Message):
    """Get Federation Options associated with run"""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    RUN_ID_FIELD_NUMBER: builtins.int
    run_id: builtins.int
    def __init__(
        self,
        *,
        run_id: builtins.int = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["run_id", b"run_id"]) -> None: ...

global___GetFederationOptionsRequest = GetFederationOptionsRequest

@typing.final
class GetFederationOptionsResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEDERATION_OPTIONS_FIELD_NUMBER: builtins.int
    @property
    def federation_options(self) -> flwr.proto.recorddict_pb2.ConfigRecord: ...
    def __init__(
        self,
        *,
        federation_options: flwr.proto.recorddict_pb2.ConfigRecord | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["federation_options", b"federation_options"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["federation_options", b"federation_options"]) -> None: ...

global___GetFederationOptionsResponse = GetFederationOptionsResponse



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2024 Flower Labs GmbH. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import typing

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class DoubleList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.float] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___DoubleList = DoubleList

@typing.final
class SintList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___SintList = SintList

@typing.final
class UintList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___UintList = UintList

@typing.final
class BoolList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.bool] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___BoolList = BoolList

@typing.final
class StringList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___StringList = StringList

@typing.final
class BytesList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALS_FIELD_NUMBER: builtins.int
    @property
    def vals(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(
        self,
        *,
        vals: collections.abc.Iterable[builtins.bytes] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["vals", b"vals"]) -> None: ...

global___BytesList = BytesList

@typing.final
class Array(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DTYPE_FIELD_NUMBER: builtins.int
    SHAPE_FIELD_NUMBER: builtins.int
    STYPE_FIELD_NUMBER: builtins.int
    DATA_FIELD_NUMBER: builtins.int
    dtype: builtins.str
    stype: builtins.str
    data: builtins.bytes
    @property
    def shape(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(
        self,
        *,
        dtype: builtins.str = ...,
        shape: collections.abc.Iterable[builtins.int] | None = ...,
        stype: builtins.str = ...,
        data: builtins.bytes = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["data", b"data", "dtype", b"dtype", "shape", b"shape", "stype", b"stype"]) -> None: ...

global___Array = Array

@typing.final
class MetricRecordValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DOUBLE_FIELD_NUMBER: builtins.int
    SINT64_FIELD_NUMBER: builtins.int
    UINT64_FIELD_NUMBER: builtins.int
    DOUBLE_LIST_FIELD_NUMBER: builtins.int
    SINT_LIST_FIELD_NUMBER: builtins.int
    UINT_LIST_FIELD_NUMBER: builtins.int
    double: builtins.float
    """Single element"""
    sint64: builtins.int
    uint64: builtins.int
    @property
    def double_list(self) -> global___DoubleList:
        """List types"""

    @property
    def sint_list(self) -> global___SintList: ...
    @property
    def uint_list(self) -> global___UintList: ...
    def __init__(
        self,
        *,
        double: builtins.float = ...,
        sint64: builtins.int = ...,
        uint64: builtins.int = ...,
        double_list: global___DoubleList | None = ...,
        sint_list: global___SintList | None = ...,
        uint_list: global___UintList | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["double", b"double", "double_list", b"double_list", "sint64", b"sint64", "sint_list", b"sint_list", "uint64", b"uint64", "uint_list", b"uint_list", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["double", b"double", "double_list", b"double_list", "sint64", b"sint64", "sint_list", b"sint_list", "uint64", b"uint64", "uint_list", b"uint_list", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["value", b"value"]) -> typing.Literal["double", "sint64", "uint64", "double_list", "sint_list", "uint_list"] | None: ...

global___MetricRecordValue = MetricRecordValue

@typing.final
class ConfigRecordValue(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DOUBLE_FIELD_NUMBER: builtins.int
    SINT64_FIELD_NUMBER: builtins.int
    UINT64_FIELD_NUMBER: builtins.int
    BOOL_FIELD_NUMBER: builtins.int
    STRING_FIELD_NUMBER: builtins.int
    BYTES_FIELD_NUMBER: builtins.int
    DOUBLE_LIST_FIELD_NUMBER: builtins.int
    SINT_LIST_FIELD_NUMBER: builtins.int
    UINT_LIST_FIELD_NUMBER: builtins.int
    BOOL_LIST_FIELD_NUMBER: builtins.int
    STRING_LIST_FIELD_NUMBER: builtins.int
    BYTES_LIST_FIELD_NUMBER: builtins.int
    double: builtins.float
    """Single element"""
    sint64: builtins.int
    uint64: builtins.int
    bool: builtins.bool
    string: builtins.str
    bytes: builtins.bytes
    @property
    def double_list(self) -> global___DoubleList:
        """List types"""

    @property
    def sint_list(self) -> global___SintList: ...
    @property
    def uint_list(self) -> global___UintList: ...
    @property
    def bool_list(self) -> global___BoolList: ...
    @property
    def string_list(self) -> global___StringList: ...
    @property
    def bytes_list(self) -> global___BytesList: ...
    def __init__(
        self,
        *,
        double: builtins.float = ...,
        sint64: builtins.int = ...,
        uint64: builtins.int = ...,
        bool: builtins.bool = ...,
        string: builtins.str = ...,
        bytes: builtins.bytes = ...,
        double_list: global___DoubleList | None = ...,
        sint_list: global___SintList | None = ...,
        uint_list: global___UintList | None = ...,
        bool_list: global___BoolList | None = ...,
        string_list: global___StringList | None = ...,
        bytes_list: global___BytesList | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["bool", b"bool", "bool_list", b"bool_list", "bytes", b"bytes", "bytes_list", b"bytes_list", "double", b"double", "double_list", b"double_list", "sint64", b"sint64", "sint_list", b"sint_list", "string", b"string", "string_list", b"string_list", "uint64", b"uint64", "uint_list", b"uint_list", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["bool", b"bool", "bool_list", b"bool_list", "bytes", b"bytes", "bytes_list", b"bytes_list", "double", b"double", "double_list", b"double_list", "sint64", b"sint64", "sint_list", b"sint_list", "string", b"string", "string_list", b"string_list", "uint64", b"uint64", "uint_list", b"uint_list", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing.Literal["value", b"value"]) -> typing.Literal["double", "sint64", "uint64", "bool", "string", "bytes", "double_list", "sint_list", "uint_list", "bool_list", "string_list", "bytes_list"] | None: ...

global___ConfigRecordValue = ConfigRecordValue

@typing.final
class ArrayRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Item(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___Array: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___Array | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ArrayRecord.Item]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___ArrayRecord.Item] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["items", b"items"]) -> None: ...

global___ArrayRecord = ArrayRecord

@typing.final
class MetricRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Item(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___MetricRecordValue: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___MetricRecordValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MetricRecord.Item]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___MetricRecord.Item] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["items", b"items"]) -> None: ...

global___MetricRecord = MetricRecord

@typing.final
class ConfigRecord(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Item(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___ConfigRecordValue: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            value: global___ConfigRecordValue | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___ConfigRecord.Item]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___ConfigRecord.Item] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["items", b"items"]) -> None: ...

global___ConfigRecord = ConfigRecord

@typing.final
class RecordDict(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class Item(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        ARRAY_RECORD_FIELD_NUMBER: builtins.int
        METRIC_RECORD_FIELD_NUMBER: builtins.int
        CONFIG_RECORD_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def array_record(self) -> global___ArrayRecord: ...
        @property
        def metric_record(self) -> global___MetricRecord: ...
        @property
        def config_record(self) -> global___ConfigRecord: ...
        def __init__(
            self,
            *,
            key: builtins.str = ...,
            array_record: global___ArrayRecord | None = ...,
            metric_record: global___MetricRecord | None = ...,
            config_record: global___ConfigRecord | None = ...,
        ) -> None: ...
        def HasField(self, field_name: typing.Literal["array_record", b"array_record", "config_record", b"config_record", "metric_record", b"metric_record", "value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["array_record", b"array_record", "config_record", b"config_record", "key", b"key", "metric_record", b"metric_record", "value", b"value"]) -> None: ...
        def WhichOneof(self, oneof_group: typing.Literal["value", b"value"]) -> typing.Literal["array_record", "metric_record", "config_record"] | None: ...

    ITEMS_FIELD_NUMBER: builtins.int
    @property
    def items(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___RecordDict.Item]: ...
    def __init__(
        self,
        *,
        items: collections.abc.Iterable[global___RecordDict.Item] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing.Literal["items", b"items"]) -> None: ...

global___RecordDict = RecordDict



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Fleet API gRPC request-response servicer."""


import threading
from logging import DEBUG, ERROR, INFO

import grpc
from google.protobuf.json_format import MessageToDict

from flwr.common.constant import PUBLIC_KEY_ALREADY_IN_USE_MESSAGE
from flwr.common.inflatable import UnexpectedObjectContentError
from flwr.common.logger import log
from flwr.common.typing import InvalidRunStatusException
from flwr.proto import fleet_pb2_grpc  # pylint: disable=E0611
from flwr.proto.fab_pb2 import GetFabRequest, GetFabResponse  # pylint: disable=E0611
from flwr.proto.fleet_pb2 import (  # pylint: disable=E0611
    ActivateNodeRequest,
    ActivateNodeResponse,
    DeactivateNodeRequest,
    DeactivateNodeResponse,
    PullMessagesRequest,
    PullMessagesResponse,
    PushMessagesRequest,
    PushMessagesResponse,
    RegisterNodeFleetRequest,
    RegisterNodeFleetResponse,
    UnregisterNodeFleetRequest,
    UnregisterNodeFleetResponse,
)
from flwr.proto.heartbeat_pb2 import (  # pylint: disable=E0611
    SendNodeHeartbeatRequest,
    SendNodeHeartbeatResponse,
)
from flwr.proto.message_pb2 import (  # pylint: disable=E0611
    ConfirmMessageReceivedRequest,
    ConfirmMessageReceivedResponse,
    PullObjectRequest,
    PullObjectResponse,
    PushObjectRequest,
    PushObjectResponse,
)
from flwr.proto.run_pb2 import GetRunRequest, GetRunResponse  # pylint: disable=E0611
from flwr.server.superlink.fleet.message_handler import message_handler
from flwr.server.superlink.linkstate import LinkStateFactory
from flwr.server.superlink.utils import abort_grpc_context
from flwr.supercore.ffs import FfsFactory
from flwr.supercore.object_store import ObjectStoreFactory


class FleetServicer(fleet_pb2_grpc.FleetServicer):
    """Fleet API servicer."""

    def __init__(
        self,
        state_factory: LinkStateFactory,
        ffs_factory: FfsFactory,
        objectstore_factory: ObjectStoreFactory,
        enable_supernode_auth: bool,
    ) -> None:
        self.state_factory = state_factory
        self.ffs_factory = ffs_factory
        self.objectstore_factory = objectstore_factory
        self.enable_supernode_auth = enable_supernode_auth
        self.lock = threading.Lock()

    def RegisterNode(
        self, request: RegisterNodeFleetRequest, context: grpc.ServicerContext
    ) -> RegisterNodeFleetResponse:
        """Register a node."""
        # Prevent registration when SuperNode authentication is enabled
        if self.enable_supernode_auth:
            log(ERROR, "SuperNode registration is disabled through Fleet API.")
            context.abort(
                grpc.StatusCode.FAILED_PRECONDITION,
                "SuperNode authentication is enabled. "
                "All SuperNodes must be registered via the CLI.",
            )

        try:
            response = message_handler.register_node(
                request=request,
                state=self.state_factory.state(),
            )
            log(DEBUG, "[Fleet.RegisterNode] Registered node_id=%s", response.node_id)
            return response
        except ValueError:
            # Public key already in use
            # This should NEVER happen due to the public keys should be automatically
            # generated and unique for each SuperNode instance.
            log(
                ERROR,
                "[Fleet.RegisterNode] Registration failed: %s",
                PUBLIC_KEY_ALREADY_IN_USE_MESSAGE,
            )
            context.abort(
                grpc.StatusCode.FAILED_PRECONDITION, PUBLIC_KEY_ALREADY_IN_USE_MESSAGE
            )

        raise RuntimeError  # Make mypy happy

    def ActivateNode(
        self, request: ActivateNodeRequest, context: grpc.ServicerContext
    ) -> ActivateNodeResponse:
        """Activate a node."""
        try:
            response = message_handler.activate_node(
                request=request,
                state=self.state_factory.state(),
            )
            log(INFO, "[Fleet.ActivateNode] Activated node_id=%s", response.node_id)
            return response
        except message_handler.InvalidHeartbeatIntervalError:
            # Heartbeat interval is invalid
            log(ERROR, "[Fleet.ActivateNode] Invalid heartbeat interval")
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT, "Invalid heartbeat interval"
            )
        except ValueError as e:
            log(ERROR, "[Fleet.ActivateNode] Activation failed: %s", str(e))
            context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))

        raise RuntimeError  # Make mypy happy

    def DeactivateNode(
        self, request: DeactivateNodeRequest, context: grpc.ServicerContext
    ) -> DeactivateNodeResponse:
        """Deactivate a node."""
        try:
            response = message_handler.deactivate_node(
                request=request,
                state=self.state_factory.state(),
            )
            log(INFO, "[Fleet.DeactivateNode] Deactivated node_id=%s", request.node_id)
            return response
        except ValueError as e:
            log(ERROR, "[Fleet.DeactivateNode] Deactivation failed: %s", str(e))
            context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))

        raise RuntimeError  # Make mypy happy

    def UnregisterNode(
        self, request: UnregisterNodeFleetRequest, context: grpc.ServicerContext
    ) -> UnregisterNodeFleetResponse:
        """Unregister a node."""
        # Prevent unregistration when SuperNode authentication is enabled
        if self.enable_supernode_auth:
            log(ERROR, "SuperNode unregistration is disabled through Fleet API.")
            context.abort(
                grpc.StatusCode.FAILED_PRECONDITION,
                "SuperNode authentication is enabled. "
                "All SuperNodes must be unregistered via the CLI.",
            )

        try:
            response = message_handler.unregister_node(
                request=request,
                state=self.state_factory.state(),
            )
            log(
                DEBUG, "[Fleet.UnregisterNode] Unregistered node_id=%s", request.node_id
            )
            return response
        except ValueError as e:
            log(
                ERROR,
                "[Fleet.UnregisterNode] Unregistration failed: %s",
                str(e),
            )
            context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))
            raise RuntimeError from None  # Make mypy happy

    def SendNodeHeartbeat(
        self, request: SendNodeHeartbeatRequest, context: grpc.ServicerContext
    ) -> SendNodeHeartbeatResponse:
        """."""
        log(DEBUG, "[Fleet.SendNodeHeartbeat] Request: %s", MessageToDict(request))
        try:
            return message_handler.send_node_heartbeat(
                request=request,
                state=self.state_factory.state(),
            )
        except message_handler.InvalidHeartbeatIntervalError:
            # Heartbeat interval is invalid
            log(ERROR, "[Fleet.SendNodeHeartbeat] Invalid heartbeat interval")
            context.abort(
                grpc.StatusCode.INVALID_ARGUMENT, "Invalid heartbeat interval"
            )
        raise RuntimeError  # Make mypy happy

    def PullMessages(
        self, request: PullMessagesRequest, context: grpc.ServicerContext
    ) -> PullMessagesResponse:
        """Pull Messages."""
        log(INFO, "[Fleet.PullMessages] node_id=%s", request.node.node_id)
        log(DEBUG, "[Fleet.PullMessages] Request: %s", MessageToDict(request))
        return message_handler.pull_messages(
            request=request,
            state=self.state_factory.state(),
            store=self.objectstore_factory.store(),
        )

    def PushMessages(
        self, request: PushMessagesRequest, context: grpc.ServicerContext
    ) -> PushMessagesResponse:
        """Push Messages."""
        if request.messages_list:
            log(
                INFO,
                "[Fleet.PushMessages] Push replies from node_id=%s",
                request.messages_list[0].metadata.src_node_id,
            )
        else:
            log(INFO, "[Fleet.PushMessages] No replies to push")

        try:
            res = message_handler.push_messages(
                request=request,
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except InvalidRunStatusException as e:
            abort_grpc_context(e.message, context)

        return res

    def GetRun(
        self, request: GetRunRequest, context: grpc.ServicerContext
    ) -> GetRunResponse:
        """Get run information."""
        log(INFO, "[Fleet.GetRun] Requesting `Run` for run_id=%s", request.run_id)

        try:
            res = message_handler.get_run(
                request=request,
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except (InvalidRunStatusException, ValueError) as e:
            abort_grpc_context(str(e), context)

        return res

    def GetFab(
        self, request: GetFabRequest, context: grpc.ServicerContext
    ) -> GetFabResponse:
        """Get FAB."""
        log(INFO, "[Fleet.GetFab] Requesting FAB for fab_hash=%s", request.hash_str)
        try:
            res = message_handler.get_fab(
                request=request,
                ffs=self.ffs_factory.ffs(),
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except (InvalidRunStatusException, ValueError) as e:
            abort_grpc_context(str(e), context)

        return res

    def PushObject(
        self, request: PushObjectRequest, context: grpc.ServicerContext
    ) -> PushObjectResponse:
        """Push an object to the ObjectStore."""
        log(
            DEBUG,
            "[Fleet.PushObject] Push Object with object_id=%s",
            request.object_id,
        )

        try:
            # Insert in Store
            res = message_handler.push_object(
                request=request,
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except InvalidRunStatusException as e:
            abort_grpc_context(e.message, context)
        except UnexpectedObjectContentError as e:
            # Object content is not valid
            context.abort(grpc.StatusCode.FAILED_PRECONDITION, str(e))

        return res

    def PullObject(
        self, request: PullObjectRequest, context: grpc.ServicerContext
    ) -> PullObjectResponse:
        """Pull an object from the ObjectStore."""
        log(
            DEBUG,
            "[Fleet.PullObject] Pull Object with object_id=%s",
            request.object_id,
        )

        try:
            # Fetch from store
            res = message_handler.pull_object(
                request=request,
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except InvalidRunStatusException as e:
            abort_grpc_context(e.message, context)

        return res

    def ConfirmMessageReceived(
        self, request: ConfirmMessageReceivedRequest, context: grpc.ServicerContext
    ) -> ConfirmMessageReceivedResponse:
        """Confirm message received."""
        log(
            DEBUG,
            "[Fleet.ConfirmMessageReceived] Message with ID '%s' has been received",
            request.message_object_id,
        )

        try:
            res = message_handler.confirm_message_received(
                request=request,
                state=self.state_factory.state(),
                store=self.objectstore_factory.store(),
            )
        except InvalidRunStatusException as e:
            abort_grpc_context(e.message, context)

        return res



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Abstract base class for ArtifactProvider."""


from abc import ABC, abstractmethod


class ArtifactProvider(ABC):
    """ArtifactProvider interface for providing artifact download links."""

    @abstractmethod
    def get_url(self, run_id: int) -> str | None:
        """Return the artifact download link for the given run ID."""

    @property
    @abstractmethod
    def output_dir(self) -> str:
        """Permanent storage directory."""

    @property
    @abstractmethod
    def tmp_dir(self) -> str:
        """Temporary storage directory."""



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: flwr/proto/federation.proto
# Protobuf Python Version: 5.29.0
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    29,
    0,
    '',
    'flwr/proto/federation.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from flwr.proto import run_pb2 as flwr_dot_proto_dot_run__pb2
from flwr.proto import node_pb2 as flwr_dot_proto_dot_node__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1b\x66lwr/proto/federation.proto\x12\nflwr.proto\x1a\x14\x66lwr/proto/run.proto\x1a\x15\x66lwr/proto/node.proto\"s\n\nFederation\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0bmember_aids\x18\x02 \x03(\t\x12#\n\x05nodes\x18\x03 \x03(\x0b\x32\x14.flwr.proto.NodeInfo\x12\x1d\n\x04runs\x18\x04 \x03(\x0b\x32\x0f.flwr.proto.Runb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'flwr.proto.federation_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_FEDERATION']._serialized_start=88
  _globals['_FEDERATION']._serialized_end=203
# @@protoc_insertion_point(module_scope)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `pull` command."""


from pathlib import Path
from typing import Annotated

import typer

from flwr.cli.config_utils import (
    exit_if_no_address,
    load_and_validate,
    process_loaded_project_config,
    validate_federation_in_project_config,
)
from flwr.cli.constant import FEDERATION_CONFIG_HELP_MESSAGE
from flwr.common.constant import FAB_CONFIG_FILE
from flwr.proto.control_pb2 import (  # pylint: disable=E0611
    PullArtifactsRequest,
    PullArtifactsResponse,
)
from flwr.proto.control_pb2_grpc import ControlStub

from .utils import flwr_cli_grpc_exc_handler, init_channel, load_cli_auth_plugin


def pull(  # pylint: disable=R0914
    run_id: Annotated[
        int,
        typer.Option(
            "--run-id",
            help="Run ID to pull artifacts from.",
        ),
    ],
    app: Annotated[
        Path,
        typer.Argument(help="Path of the Flower App to run."),
    ] = Path("."),
    federation: Annotated[
        str | None,
        typer.Argument(help="Name of the federation."),
    ] = None,
    federation_config_overrides: Annotated[
        list[str] | None,
        typer.Option(
            "--federation-config",
            help=FEDERATION_CONFIG_HELP_MESSAGE,
        ),
    ] = None,
) -> None:
    """Pull artifacts from a Flower run.

    Retrieve a download URL for artifacts generated during a completed Flower run. The
    artifacts can then be downloaded from the provided URL.
    """
    typer.secho("Loading project configuration... ", fg=typer.colors.BLUE)

    pyproject_path = app / FAB_CONFIG_FILE if app else None
    config, errors, warnings = load_and_validate(pyproject_path, check_module=False)
    config = process_loaded_project_config(config, errors, warnings)
    federation, federation_config = validate_federation_in_project_config(
        federation, config, federation_config_overrides
    )
    exit_if_no_address(federation_config, "pull")
    channel = None
    try:

        auth_plugin = load_cli_auth_plugin(app, federation, federation_config)
        channel = init_channel(app, federation_config, auth_plugin)
        stub = ControlStub(channel)
        with flwr_cli_grpc_exc_handler():
            res: PullArtifactsResponse = stub.PullArtifacts(
                PullArtifactsRequest(run_id=run_id)
            )

        if not res.url:
            typer.secho(
                f"❌ A download URL for artifacts from run {run_id} couldn't be "
                "obtained.",
                fg=typer.colors.RED,
                bold=True,
                err=True,
            )
            raise typer.Exit(code=1)

        typer.secho(
            f"✅ Artifacts for run {run_id} can be downloaded from: {res.url}",
            fg=typer.colors.GREEN,
        )
    finally:
        if channel:
            channel.close()



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Flower command line interface `app review` command."""


import base64
import hashlib
import re
from pathlib import Path
from typing import Annotated

import requests
import typer
from cryptography.exceptions import UnsupportedAlgorithm
from cryptography.hazmat.primitives.asymmetric import ed25519

from flwr.common import now
from flwr.common.config import get_flwr_dir
from flwr.common.constant import FAB_CONFIG_FILE
from flwr.common.version import package_version as flwr_version
from flwr.supercore.constant import PLATFORM_API_URL
from flwr.supercore.primitives.asymmetric_ed25519 import (
    create_message_to_sign,
    load_private_key,
    sign_message,
)

from ..auth_plugin.oidc_cli_plugin import OidcCliPlugin
from ..config_utils import (
    load_and_validate,
    process_loaded_project_config,
    validate_federation_in_project_config,
)
from ..constant import FEDERATION_CONFIG_HELP_MESSAGE
from ..install import install_from_fab
from ..utils import load_cli_auth_plugin, parse_app_spec, request_download_link

TRY_AGAIN_MESSAGE = "Please try again or press CTRL+C to abort.\n"


# pylint: disable-next=too-many-locals, too-many-statements
def review(
    app_spec: Annotated[
        str,
        typer.Argument(
            help="App specifier (e.g., '@account/app' or '@account/app==1.0.0'). "
            "Version is optional; defaults to the latest."
        ),
    ],
    app_dir_login: Annotated[
        Path,
        typer.Argument(
            help="Project directory to used for login before reviewing app."
        ),
    ] = Path("."),
    federation: Annotated[
        str | None,
        typer.Argument(
            help="Name of the federation used for login before reviewing app."
        ),
    ] = None,
    federation_config_overrides: Annotated[
        list[str] | None,
        typer.Option(
            "--federation-config",
            help=FEDERATION_CONFIG_HELP_MESSAGE,
        ),
    ] = None,
) -> None:
    """Download a FAB for <APP-ID>, unpack it for manual review, and upon confirmation
    sign & submit the review to the Platform."""
    # Load configs
    pyproject_path = app_dir_login / FAB_CONFIG_FILE if app_dir_login else None
    config, errors, warnings = load_and_validate(pyproject_path, check_module=False)
    config = process_loaded_project_config(config, errors, warnings)
    federation, federation_config = validate_federation_in_project_config(
        federation, config, federation_config_overrides
    )

    # Load the authentication plugin
    auth_plugin = load_cli_auth_plugin(app_dir_login, federation, federation_config)
    auth_plugin.load_tokens()
    if not isinstance(auth_plugin, OidcCliPlugin) or not auth_plugin.access_token:
        typer.secho(
            "❌ Please log in before reviewing app.",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1)

    # Load token from the plugin
    token = auth_plugin.access_token

    # Validate app version and ID format
    app_id, app_version = parse_app_spec(app_spec)

    # Download FAB
    typer.secho("Downloading FAB... ", fg=typer.colors.BLUE)
    url = f"{PLATFORM_API_URL}/hub/fetch-fab"
    presigned_url = request_download_link(app_id, app_version, url, "fab_url")
    fab_bytes = _download_fab(presigned_url)

    # Unpack FAB
    typer.secho("Unpacking FAB... ", fg=typer.colors.BLUE)
    review_dir = _create_review_dir()
    review_app_path = install_from_fab(fab_bytes, review_dir)

    # Extract app version
    version_pattern = re.compile(r"\b(\d+\.\d+\.\d+)\b")
    match = version_pattern.search(str(review_app_path))
    assert match is not None
    app_version = match.group(1)

    # Prompt to ask for sign
    typer.secho(
        f"""
    Review the unpacked app in the following directory:

        {typer.style(review_app_path, fg=typer.colors.GREEN, bold=True)}

    If you have reviewed the app and want to continue to sign it,
    type {typer.style("SIGN", fg=typer.colors.GREEN, bold=True)} or abort with CTRL+C.
    """,
        fg=typer.colors.BLUE,
    )

    confirmation = typer.prompt("Type SIGN to continue").strip()
    if confirmation.upper() != "SIGN":
        typer.secho("Aborted (user did not type SIGN).", fg=typer.colors.YELLOW)
        raise typer.Exit(code=130)

    # Ask for private key path (retry until valid)
    while True:
        try:
            key_path_str = typer.prompt(
                "Please specify the path of Ed25519 OpenSSH private key for signing"
            )
        except typer.Abort as e:
            typer.secho("Aborted by user.", fg=typer.colors.YELLOW, err=True)
            raise typer.Exit(code=130) from e

        key_path = Path(key_path_str).expanduser().resolve()

        if not key_path.is_file():
            typer.secho(
                f"❌ Private key not found: {key_path}",
                fg=typer.colors.RED,
                err=True,
            )
            typer.secho(TRY_AGAIN_MESSAGE, fg=typer.colors.YELLOW)
            continue

        # Load private key
        try:
            private_key = load_private_key(key_path)
        except (OSError, ValueError, UnsupportedAlgorithm) as e:
            typer.secho(
                f"❌ Failed to load the private key: {e}", fg=typer.colors.RED, err=True
            )
            typer.secho(TRY_AGAIN_MESSAGE, fg=typer.colors.YELLOW)
            continue
        break  # valid

    # Sign FAB
    signature, signed_at = _sign_fab(fab_bytes, private_key)

    # Submit review
    _submit_review(app_id, app_version, signature, signed_at, token)


def _create_review_dir() -> Path:
    """Create a directory for reviewing code."""
    home = get_flwr_dir()
    review_dir = home / "reviews"
    review_dir.mkdir(parents=True, exist_ok=True)
    return review_dir


def _download_fab(url: str) -> bytes:
    """Download FAB file from given URL."""
    try:
        r = requests.get(url, timeout=60)
        r.raise_for_status()
    except requests.RequestException as e:
        typer.secho(
            f"❌ FAB download failed: {e}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1) from e
    return r.content


def _sign_fab(
    fab_bytes: bytes, private_key: ed25519.Ed25519PrivateKey
) -> tuple[bytes, int]:
    """Sign the given FAB hash bytes."""
    # Get current timestamp
    timestamp = int(now().timestamp())
    message_to_sign = create_message_to_sign(
        hashlib.sha256(fab_bytes).digest(),
        timestamp,
    )
    return sign_message(private_key, message_to_sign), timestamp


def _submit_review(
    app_id: str, app_version: str, signature: bytes, signed_at: int, token: str
) -> None:
    """Submit review to Flower Platform API."""
    signature_b64 = base64.urlsafe_b64encode(signature).rstrip(b"=").decode("ascii")
    url = f"{PLATFORM_API_URL}/hub/apps/signature"
    headers = {"Authorization": f"Bearer {token}", "Content-Type": "application/json"}
    payload = {
        "app_id": app_id,
        "app_version": app_version,
        "signature_b64": signature_b64,
        "signed_at": signed_at,
        "flwr_version": flwr_version,
    }
    try:
        resp = requests.post(url, headers=headers, json=payload, timeout=120)
    except requests.RequestException as e:
        typer.secho(
            f"❌ Network error while submitting review: {e}",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(code=1) from e

    if resp.ok:
        typer.secho("🎊 Review submitted", fg=typer.colors.GREEN, bold=True)
        return

    # Error path:
    msg = f"❌ Review submission failed (HTTP {resp.status_code})"
    if resp.text:
        msg += f": {resp.text}"
    typer.secho(msg, fg=typer.colors.RED, err=True)
    raise typer.Exit(code=1)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Federated Averaging with Trimmed Mean [Dong Yin, et al., 2021].

Paper: arxiv.org/abs/1803.01498
"""


from collections.abc import Callable, Iterable
from logging import INFO
from typing import cast

import numpy as np

from flwr.common import Array, ArrayRecord, Message, MetricRecord, NDArray, RecordDict
from flwr.common.logger import log

from ..exception import AggregationError
from .fedavg import FedAvg


class FedTrimmedAvg(FedAvg):
    """Federated Averaging with Trimmed Mean [Dong Yin, et al., 2021].

    Implemented based on: https://arxiv.org/abs/1803.01498

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for both ArrayRecords and MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    beta : float (default: 0.2)
        Fraction to cut off of both tails of the distribution.
    """

    def __init__(  # pylint: disable=R0913, R0917
        self,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_train_nodes: int = 2,
        min_evaluate_nodes: int = 2,
        min_available_nodes: int = 2,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        beta: float = 0.2,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=min_train_nodes,
            min_evaluate_nodes=min_evaluate_nodes,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
        )
        self.beta = beta

    def summary(self) -> None:
        """Log summary configuration of the strategy."""
        log(INFO, "\t├──> FedTrimmedAvg settings:")
        log(INFO, "\t│\t└── beta: %s", self.beta)
        super().summary()

    def aggregate_train(
        self,
        server_round: int,
        replies: Iterable[Message],
    ) -> tuple[ArrayRecord | None, MetricRecord | None]:
        """Aggregate ArrayRecords and MetricRecords in the received Messages."""
        # Call FedAvg aggregate_train to perform validation and aggregation
        valid_replies, _ = self._check_and_log_replies(replies, is_train=True)

        if not valid_replies:
            return None, None

        # Aggregate ArrayRecords using trimmed mean
        # Get the key for the only ArrayRecord from the first Message
        record_key = list(valid_replies[0].content.array_records.keys())[0]
        # Preserve keys for arrays in ArrayRecord
        array_keys = list(valid_replies[0].content[record_key].keys())

        # Compute trimmed mean for each layer and construct ArrayRecord
        arrays = ArrayRecord()
        for array_key in array_keys:
            # Get the corresponding layer from each client
            layers = [
                cast(ArrayRecord, msg.content[record_key]).pop(array_key).numpy()
                for msg in valid_replies
            ]
            # Compute trimmed mean and save as Array in ArrayRecord
            try:
                arrays[array_key] = Array(trim_mean(np.stack(layers), self.beta))
            except ValueError as e:
                raise AggregationError(
                    f"Trimmed mean could not be computed. "
                    f"Likely cause: beta={self.beta} is too large."
                ) from e

        # Aggregate MetricRecords
        metrics = self.train_metrics_aggr_fn(
            [msg.content for msg in valid_replies],
            self.weighted_by_key,
        )
        return arrays, metrics


def trim_mean(array: NDArray, cut_fraction: float) -> NDArray:
    """Compute trimmed mean along axis=0.

    It is based on the scipy implementation:

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.trim_mean.html
    """
    axis = 0
    nobs = array.shape[0]
    lowercut = int(cut_fraction * nobs)
    uppercut = nobs - lowercut
    if lowercut > uppercut:
        raise ValueError("Fraction too big.")

    atmp = np.partition(array, (lowercut, uppercut - 1), axis)

    slice_list = [slice(None)] * atmp.ndim
    slice_list[axis] = slice(lowercut, uppercut)
    result: NDArray = np.mean(atmp[tuple(slice_list)], axis=axis)
    return result



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Federated Optimization (FedProx) [Li et al., 2018] strategy.

Paper: arxiv.org/abs/1812.06127
"""


from collections.abc import Callable, Iterable
from logging import INFO, WARN

from flwr.common import (
    ArrayRecord,
    ConfigRecord,
    Message,
    MetricRecord,
    RecordDict,
    log,
)
from flwr.server import Grid

from .fedavg import FedAvg


class FedProx(FedAvg):
    r"""Federated Optimization strategy.

    Implementation based on https://arxiv.org/abs/1812.06127

    FedProx extends FedAvg by introducing a proximal term into the client-side
    optimization objective. The strategy itself behaves identically to FedAvg
    on the server side, but each client **MUST** add a proximal regularization
    term to its local loss function during training:

    .. math::
        \frac{\mu}{2} || w - w^t ||^2

    Where $w^t$ denotes the global parameters and $w$ denotes the local weights
    being optimized.

    This strategy sends the proximal term inside the ``ConfigRecord`` as part of the
    ``configure_train`` method under key ``"proximal-mu"``. The client can then use this
    value to add the proximal term to the loss function.

    In PyTorch, for example, the loss would go from:

    .. code:: python
        loss = criterion(net(inputs), labels)

    To:

    .. code:: python
        # Get proximal term weight from message
        mu = msg.content["config"]["proximal-mu"]

        # Compute proximal term
        proximal_term = 0.0
        for local_weights, global_weights in zip(net.parameters(), global_params):
            proximal_term += (local_weights - global_weights).norm(2)

        # Update loss
        loss = criterion(net(inputs), labels) + (mu / 2) * proximal_term

    With ``global_params`` being a copy of the model parameters, created **after**
    applying the received global weights but **before** local training begins.

    .. code:: python
        global_params = copy.deepcopy(net).parameters()

    Parameters
    ----------
    fraction_train : float (default: 1.0)
        Fraction of nodes used during training. In case `min_train_nodes`
        is larger than `fraction_train * total_connected_nodes`, `min_train_nodes`
        will still be sampled.
    fraction_evaluate : float (default: 1.0)
        Fraction of nodes used during validation. In case `min_evaluate_nodes`
        is larger than `fraction_evaluate * total_connected_nodes`,
        `min_evaluate_nodes` will still be sampled.
    min_train_nodes : int (default: 2)
        Minimum number of nodes used during training.
    min_evaluate_nodes : int (default: 2)
        Minimum number of nodes used during validation.
    min_available_nodes : int (default: 2)
        Minimum number of total nodes in the system.
    weighted_by_key : str (default: "num-examples")
        The key within each MetricRecord whose value is used as the weight when
        computing weighted averages for both ArrayRecords and MetricRecords.
    arrayrecord_key : str (default: "arrays")
        Key used to store the ArrayRecord when constructing Messages.
    configrecord_key : str (default: "config")
        Key used to store the ConfigRecord when constructing Messages.
    train_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    evaluate_metrics_aggr_fn : Optional[callable] (default: None)
        Function with signature (list[RecordDict], str) -> MetricRecord,
        used to aggregate MetricRecords from training round replies.
        If `None`, defaults to `aggregate_metricrecords`, which performs a weighted
        average using the provided weight factor key.
    proximal_mu : float (default: 0.0)
        The weight of the proximal term used in the optimization. 0.0 makes
        this strategy equivalent to FedAvg, and the higher the coefficient, the more
        regularization will be used (that is, the client parameters will need to be
        closer to the server parameters during training).
    """

    def __init__(  # pylint: disable=R0913, R0917
        self,
        fraction_train: float = 1.0,
        fraction_evaluate: float = 1.0,
        min_train_nodes: int = 2,
        min_evaluate_nodes: int = 2,
        min_available_nodes: int = 2,
        weighted_by_key: str = "num-examples",
        arrayrecord_key: str = "arrays",
        configrecord_key: str = "config",
        train_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        evaluate_metrics_aggr_fn: (
            Callable[[list[RecordDict], str], MetricRecord] | None
        ) = None,
        proximal_mu: float = 0.0,
    ) -> None:
        super().__init__(
            fraction_train=fraction_train,
            fraction_evaluate=fraction_evaluate,
            min_train_nodes=min_train_nodes,
            min_evaluate_nodes=min_evaluate_nodes,
            min_available_nodes=min_available_nodes,
            weighted_by_key=weighted_by_key,
            arrayrecord_key=arrayrecord_key,
            configrecord_key=configrecord_key,
            train_metrics_aggr_fn=train_metrics_aggr_fn,
            evaluate_metrics_aggr_fn=evaluate_metrics_aggr_fn,
        )
        self.proximal_mu = proximal_mu

        if self.proximal_mu == 0.0:
            log(
                WARN,
                "FedProx initialized with `proximal_mu=0.0`. "
                "This makes the strategy equivalent to FedAvg.",
            )

    def summary(self) -> None:
        """Log summary configuration of the strategy."""
        log(INFO, "\t├──> FedProx settings:")
        log(INFO, "\t│\t└── Proximal mu: %s", self.proximal_mu)
        super().summary()

    def configure_train(
        self, server_round: int, arrays: ArrayRecord, config: ConfigRecord, grid: Grid
    ) -> Iterable[Message]:
        """Configure the next round of federated training."""
        # Inject proximal term weight into config
        config["proximal-mu"] = self.proximal_mu
        return super().configure_train(server_round, arrays, config, grid)



# Copyright 2025 Flower Labs GmbH. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Asymmetric cryptography utilities."""


from typing import cast

from cryptography.exceptions import InvalidSignature
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import ec


def generate_key_pairs() -> (
    tuple[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey]
):
    """Generate private and public key pairs with Cryptography."""
    private_key = ec.generate_private_key(ec.SECP384R1())
    public_key = private_key.public_key()
    return private_key, public_key


def private_key_to_bytes(private_key: ec.EllipticCurvePrivateKey) -> bytes:
    """Serialize private key to bytes."""
    return private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )


def bytes_to_private_key(private_key_bytes: bytes) -> ec.EllipticCurvePrivateKey:
    """Deserialize private key from bytes."""
    return cast(
        ec.EllipticCurvePrivateKey,
        serialization.load_pem_private_key(data=private_key_bytes, password=None),
    )


def public_key_to_bytes(public_key: ec.EllipticCurvePublicKey) -> bytes:
    """Serialize public key to bytes."""
    return public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )


def bytes_to_public_key(public_key_bytes: bytes) -> ec.EllipticCurvePublicKey:
    """Deserialize public key from bytes."""
    return cast(
        ec.EllipticCurvePublicKey,
        serialization.load_pem_public_key(data=public_key_bytes),
    )


def sign_message(private_key: ec.EllipticCurvePrivateKey, message: bytes) -> bytes:
    """Sign a message using the provided EC private key.

    Parameters
    ----------
    private_key : ec.EllipticCurvePrivateKey
        The EC private key to sign the message with.
    message : bytes
        The message to be signed.

    Returns
    -------
    bytes
        The signature of the message.
    """
    signature = private_key.sign(message, ec.ECDSA(hashes.SHA256()))
    return signature


def verify_signature(
    public_key: ec.EllipticCurvePublicKey, message: bytes, signature: bytes
) -> bool:
    """Verify a signature against a message using the provided EC public key.

    Parameters
    ----------
    public_key : ec.EllipticCurvePublicKey
        The EC public key to verify the signature.
    message : bytes
        The original message.
    signature : bytes
        The signature to verify.

    Returns
    -------
    bool
        True if the signature is valid, False otherwise.
    """
    try:
        public_key.verify(signature, message, ec.ECDSA(hashes.SHA256()))
        return True
    except InvalidSignature:
        return False


def uses_nist_ec_curve(public_key: ec.EllipticCurvePublicKey) -> bool:
    """Return True if the provided key uses a NIST EC curve."""
    return isinstance(
        public_key.curve,
        (ec.SECP192R1 | ec.SECP224R1 | ec.SECP256R1 | ec.SECP384R1 | ec.SECP521R1),
    )
